const zookeeper = require('node-zookeeper-client');
const CreateMode = zookeeper.CreateMode;

module.exports = function (config, log) {

    const client = zookeeper.createClient(config.connStr);

    client.on('connectedReadOnly', function () {
        log.debug('Zookeeper connected read-only');
    });

    client.on('expired', function () {
        log.debug('Zookeeper session expired');
    });

    client.on('disconnected', function () {
        log.debug('Zookeeper disconnected');
    });

    client.on('state', function (state) {
        log.debug('Zookeeper state : %s', state);
    });

    client.once('connected', function () {
        log.debug('Connected to Zookeeper at', config.connStr);
        init();
    });

    client.on('error', function (error) {
        log.debug('client error %s', error);
    });

    function promisify(func) {
        return function (...args) {
            return new Promise((resolve, reject) => {
                client[func](...args, (error, result) => {
                    if (error) {
                        reject(error);
                    } else {
                        resolve(result);
                    }
                });
            });
        };
    }

    const promisedClient = {
        mkdirp: promisify('mkdirp'),
        exists: promisify('exists'),
        remove: promisify('remove'),
        create: promisify('create'),
        setData: promisify('setData'),
        getData: promisify('getData'),
        getChildren: promisify('getChildren'),
    };

    function init() {
        log.debug('init zookeeper');
    }

    function connect() {
        client.connect();
    }

    function getChildren(root) {
        return promisedClient.getData(root);
    }

    // Return all the leaf-node paths under root, i.e. all the deepest leaf-paths
    function getLeafNodes(root) {
        return promisedClient.getChildren(root)
            .then((children) => {
                if (children && children.length > 0) {
                    return Promise.all(children.map((child) => getLeafNodes(root + '/' + child)))
                        .then((values) => {
                            // Flatten the resulting array, since the depth may vary
                            const paths = [];
                            for (let value of [].concat(...values)) {
                                if (value) {
                                    paths.push(value);
                                }
                            }
                            log.debug('got paths', paths);
                            return paths;
                        });
                }
                return root;
            });
    }

    function createDir(path) {
        log.debug('creating dir', path);
        return promisedClient.mkdirp(path)
            .then(() => {
                log.debug('created', path);
            });
    }

    function createNode(path, data) {
        return promisedClient.exists(path)
            .then((stat) => {
                if (stat) {
                    log.debug('updating', path);
                    return promisedClient.setData(path, toBytes(data));
                }
                log.debug('creating', path);
                return promisedClient.create(path, toBytes(data));
            });
    }

    function createEphemeralNode(path, data) {
        return promisedClient.exists(path)
            .then((stat) => {
                if (stat) {
                    log.debug('removing ephemeral', path);
                    return promisedClient.remove(path);
                }
            })
            .then(() => {
                log.debug('creating ephemeral', path);
                return promisedClient.create(path, toBytes(data), CreateMode.EPHEMERAL);
            });
    }


    function getData(path) {
        return promisedClient.exists(path)
            .then((stat) => {
                if (stat) {
                    return promisedClient.getData(path)
                        .then((buf) => {
                            if (buf && buf.length > 0) {
                                log.debug('path =', path, '; data =', buf.toString());
                                return buf;
                            }
                            log.debug('path =', path, '; data = no data');
                            return undefined;
                        });
                }
                log.debug('path =', path, '; data = no data');
                return undefined;
            });
    }

    function close() {
        if (client.state === zookeeper.State.DISCONNECTED) {
            log.debug('Zookeeper client is not connected');
            return Promise.resolve();
        }
        client.close();
    }

    function toBytes(obj) {
        return Buffer.from(JSON.stringify(obj));
    }


    return {
        connect: connect,
        getChildren: getChildren,
        getLeafNodes: getLeafNodes,
        createDir: createDir,
        createNode: createNode,
        createEphemeralNode: createEphemeralNode,
        getData: getData,
        close: close,
    };

};

